// PostFX.cpp : Defines the exported functions for the DLL application.
//

#include "stdafx.h"
#include "BAM.h"

void initPostFX(HDC hDC);
void exitPostFX();
void doPostFX();

const char *YesNo_txt[] = {
	"No",
	"Yes"
};

struct Ccfg {
	double gamma;
	double BlurScale;
	double BlurStrength; 
	int Enabled;
} cfg = {5.0, 1.0, 0.6, 0};

HMODULE dll_hModule;
HDC hDC;
bool DoInit = false;

BOOL APIENTRY DllMain( HMODULE hModule,
	DWORD  ul_reason_for_call,
	LPVOID lpReserved
	)
{
	dll_hModule = hModule;

	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
	}
	return TRUE;
}

LPVOID GetResource(const char *resData, const char *resType = "TEXT", DWORD *outLen = NULL)
{
	HRSRC hResource = FindResourceA(dll_hModule, resData, resType);
	if (!hResource)
		return NULL;

	HGLOBAL hLoadedResource = LoadResource(dll_hModule, hResource);
	if (!hLoadedResource)
		return NULL;

	LPVOID pLockedResource = LockResource(hLoadedResource);
	if (!pLockedResource)
		return NULL;

	if (outLen != NULL) 
		*outLen = SizeofResource(dll_hModule, hResource);

	return pLockedResource;
}

//=============================================================================
// WINGDIAPI BOOL  WINAPI SwapBuffers(HDC);
// Detour of SwapBuffers
BOOL (WINAPI *Real_SwapBuffers)(HDC) = SwapBuffers;

BOOL WINAPI Routed_SwapBuffers(HDC hDC)
{
	if (cfg.Enabled ) {
//		glClear(GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
//		glDisable(GL_STENCIL_TEST);

		if (DoInit) {
			initPostFX(hDC);
			DoInit = false;
		}

		doPostFX();
	}
	return Real_SwapBuffers(hDC);
}

//=============================================================================

extern "C" {
	// called on FP load (before any hooks to opengl is made)
	// It must exist
	// If not exist => dll is not BAM plugin and will be unloaded from RAM
	BAMEXPORT int BAM_load(HMODULE bam_module)
	{
		// BAM plugin ID - 0x4001 - PostFX
		// ID > 0x4000 - No Tracking Plugin
		// - No Tracking Plugin can't be selected as "mode", but menu is accesible in "Config"->"Plugins"
		BAM::Init( 0x4001, bam_module ); 
		BAM::menu::create("PostFX", "Bloom effect"); // second param will not be displayed for non-tracking plugins
		BAM::menu::info("#c777##-Bloom effect"DEFIW);
		BAM::menu::paramSwtich("#-Enabled:"DEFPW"#+ #cfff#%s", &cfg.Enabled, YesNo_txt, ARRAY_ENTRIES(YesNo_txt), "");

		BAM::menu::paramValue("#-Gamma:"DEFPW2"%.2f", &cfg.gamma, 0.01, 0.2, "");
		BAM::menu::paramValue("#-BlurScale:"DEFPW2"%.2f", &cfg.BlurScale, 0.05, 1.0, "");
		BAM::menu::paramValue("#-BlurStrength:"DEFPW2"%.2f", &cfg.BlurStrength, 0.05, 1.0, "");

		BAM::helpers::load("PostFX", &cfg, sizeof(cfg));

		// in this proce only menu can be created and config files can be loaded
		// now there is not opengl context ready
		return 0;
	}

	// For "Tracking" plugins, it is called when user select plugin (signal for example to create working thread and take rosurces like cams)
	// For "Non-Tracking" plugins it is called on game start
	BAMEXPORT void BAM_PluginStart()
	{
		DoInit = true;
	}

	// For "Treacking" plugins, it is called then plugin is "unselected" as mode. It is signal to release resources (like cams, one cam can be used be two or more plugins)
	// For "Non-Tracking" plugins, it is called at end of game
	BAMEXPORT void BAM_PluginStop()
	{
		exitPostFX();
	}

	// called to draw (write) additional info in BAM hud
	// if not exist, it will not make any problem
	BAMEXPORT void BAM_hudDisplay()
	{
		if (!cfg.Enabled)
			return;

		double t = double(timeGetTime() % 1000)/1000.0;
		if (t > 0.5)
			t = 1.0-t;

		// position x,y is in "virtual" hud space (1920x1080) - it is independent of res or orientation.
		// pos 0,0 is left/bottom. pos 1920,1080 is right/(1/3) or height of display 1080/1920 so it is 1080x640
		BAM::menu::printf(0, 0, 0.8+t*3, true, "#cff0##+P#-ost#+FX");
	}

	// create detours for system calls (opengl)
	BAMEXPORT void BAM_AttachDetours()
	{
		// attach SwapBuffers
		BAM::detours::Attach(&(PVOID&)Real_SwapBuffers, Routed_SwapBuffers);
	}

	// destroy detours
	BAMEXPORT void BAM_DetachDetours()
	{
		// detach SwapBuffers
		BAM::detours::Detach(&(PVOID&)Real_SwapBuffers, Routed_SwapBuffers);
	}
}

//#################################################################################################
	class wglGetProcAddressList{
		static std::vector<const char *> functionNames;
		static std::vector<void *> functionPointers;

	public:
		wglGetProcAddressList(const char *functionName, void *functionPointer)
		{
			functionNames.push_back(functionName);
			functionPointers.push_back(functionPointer);
		}

		static void Hook()
		{
			for (unsigned int i=0; i<functionNames.size(); ++i)
				*(PROC *)functionPointers[i] = wglGetProcAddress(functionNames[i]);
		}
	};
	std::vector<const char *> wglGetProcAddressList::functionNames;
	std::vector<void *> wglGetProcAddressList::functionPointers;

#define WGLGETPROCADDRESS(a,b) a p##b; wglGetProcAddressList list##b(#b,&p##b)

	// attaching to wgl functions
	WGLGETPROCADDRESS (PFNGLUSEPROGRAMPROC, glUseProgram);
	WGLGETPROCADDRESS (PFNGLUNIFORM1FPROC, glUniform1f);
	WGLGETPROCADDRESS (PFNGLUNIFORM1IPROC, glUniform1i);
	WGLGETPROCADDRESS (PFNGLGETUNIFORMLOCATIONPROC, glGetUniformLocation);
	WGLGETPROCADDRESS (PFNGLACTIVETEXTUREARBPROC, glActiveTexture);

//#################################################################################################


LONG scr_width, scr_height;
GLuint textures[2];
GLuint step1 = 0, step2 = 0, step3 = 0;

GLuint createShaders(const char *vertexSrc, const char *fragmentSrc) 
{
	DWORD olen;

	// we have to ad '\0' and end of string, so we create tmp with size+1 and add '\0' at end
	const char * vs = (const char *)GetResource(vertexSrc, "TEXT", &olen);
	char *tmpVS = new char[olen+1];
	memcpy_s(tmpVS, olen+1, vs, olen);
	tmpVS[olen] = 0;

	const char * fs = (const char *)GetResource(fragmentSrc, "TEXT", &olen);
	char *tmpFS = new char[olen+1];
	memcpy_s(tmpFS, olen+1, fs, olen);
	tmpFS[olen] = 0;

	GLuint ret = BAM::render::CompileShader( tmpVS, tmpFS );

	delete tmpVS;
	delete tmpFS;

	return ret; 
}

void initPostFX(HDC hDC) 
{
	RECT rc = {0};
	HWND hWnd = ::WindowFromDC(hDC );
	::GetClientRect( hWnd, &rc );
	LONG w = rc.left-rc.right, h = rc.top - rc.bottom;
	if (w < 0) w = -w;
	if (h < 0) h = -h;
	scr_width = w;
	scr_height = h;

	// create textures
	glGenTextures(2, textures);
	for (int i=0; i<2; ++i) {
		glBindTexture(GL_TEXTURE_2D, textures[i]);
		if (i==0)
			glTexImage2D(GL_TEXTURE_2D, 0,GL_RGBA, w, h, 0,GL_RGBA, GL_UNSIGNED_BYTE, 0);
		else
			glTexImage2D(GL_TEXTURE_2D, 0,GL_RGBA, w/2, h/2, 0,GL_RGBA, GL_UNSIGNED_BYTE, 0);

		// filtering
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); 
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	}

	glBindTexture(GL_TEXTURE_2D, 0);
	wglGetProcAddressList::Hook();
	step1 = createShaders("VS", "step1FS");
	step2 = createShaders("step2VS","step2FS");
	step3 = createShaders("VS","step3FS");
}

void exitPostFX()
{
	step1 = 0;
//dbg::Message("glDeleteTextures: %d",__LINE__);
	glDeleteTextures(2, textures);
	// shader programs are deleted by BAM.dll

	BAM::helpers::save("PostFX", &cfg, sizeof(cfg));
}

void doPostFX()
{	
	if (step1 == 0)
		return;

	// saving config
	static Ccfg oldcfg = { -1 };
	if( memcmp(&oldcfg, &cfg, sizeof(cfg)) != 0 )	{
		BAM::helpers::save("PostFX", &cfg, sizeof(cfg));
		memcpy(&oldcfg, &cfg, sizeof(cfg));
	}

	GLint viewport[4];
	glGetIntegerv( GL_VIEWPORT, viewport );
//	BAM::dbg::hudDebug("[%dx%d - %dx%d]\n", viewport[0], viewport[1], viewport[2], viewport[3]);
	DWORD w = scr_width, h = scr_height;

	glBindTexture(GL_TEXTURE_2D, textures[0]);
	glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, viewport[0], viewport[1], viewport[2], viewport[3], 0);

	glDisable(GL_ALPHA_TEST);
	glDisable(GL_DEPTH_TEST);

	GLfloat sc = 0.5;

	// selecting of "bright" pixels - gamma
	pglUseProgram(step1);

	pglUniform1f(pglGetUniformLocation(step1, "pfGamma"), (GLfloat)cfg.gamma);
	pglUniform1f(pglGetUniformLocation(step2, "texelX"), (GLfloat)sc/w);
	pglUniform1f(pglGetUniformLocation(step2, "texelY"), (GLfloat)sc/h);

	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0);
	glVertex3f(-1,-1,0);

	glTexCoord2f(1.0, 0.0);
	glVertex3f(-1+2*sc,-1.0,0);

	glTexCoord2f(1.0, 1.0);
	glVertex3f(-1+2*sc,-1+2*sc,0);

	glTexCoord2f(0.0, 1.0);
	glVertex3f(-1,-1+2*sc,0);
	glEnd();

	
	glBindTexture(GL_TEXTURE_2D, textures[1]);
	glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, viewport[0], viewport[1], GLsizei(sc*viewport[2]), GLsizei(sc*viewport[3]), 0);

	// blur
	pglUseProgram(step2);

	pglUniform1i(pglGetUniformLocation(step2, "Orientation"), 1);
	pglUniform1i(pglGetUniformLocation(step2, "BlurAmount"), 5);
	pglUniform1f(pglGetUniformLocation(step2, "BlurScale"), (GLfloat)cfg.BlurScale);
	pglUniform1f(pglGetUniformLocation(step2, "BlurStrength"), (GLfloat)cfg.BlurStrength);
	pglUniform1f(pglGetUniformLocation(step2, "TexelSize"), (GLfloat)1.0f/w);

	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0);
	glVertex3f(-1,-1,0);

	glTexCoord2f(1, 0.0);
	glVertex3f(-1+2*sc,-1.0,0);

	glTexCoord2f(1, 1);
	glVertex3f(-1+2*sc,-1+2*sc,0);

	glTexCoord2f(0.0, 1);
	glVertex3f(-1,-1+2*sc,0);
	glEnd();

	glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, viewport[0], viewport[1], GLsizei(sc*viewport[2]), GLsizei(sc*viewport[3]), 0);

	pglUniform1i(pglGetUniformLocation(step2, "Orientation"), 0);

	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0);
	glVertex3f(-1,-1,0);

	glTexCoord2f(1, 0.0);
	glVertex3f(-1+2*sc,-1.0,0);

	glTexCoord2f(1, 1);
	glVertex3f(-1+2*sc,-1+2*sc,0);

	glTexCoord2f(0.0, 1);
	glVertex3f(-1,-1+2*sc,0);
	glEnd();

	glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, viewport[0], viewport[1], GLsizei(sc*viewport[2]), GLsizei(sc*viewport[3]), 0);

	// final step: combine with org image
	pglActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, textures[0]);

	pglActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, textures[1]);

	pglUseProgram(step3);

	pglUniform1i(pglGetUniformLocation(step3, "tex"), 0);
	pglUniform1i(pglGetUniformLocation(step3, "tex2"), 1);

	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0);
	glVertex3f(-1,-1,0);

	glTexCoord2f(1.0, 0.0);
	glVertex3f(1,-1,0);

	glTexCoord2f(1.0, 1.0);
	glVertex3f(1,1,0);

	glTexCoord2f(0.0, 1.0);
	glVertex3f(-1,1,0);
	glEnd();

	pglUseProgram(0);
	pglActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, 0);
	glDisable(GL_TEXTURE_2D);
	pglActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, 0);
	glDisable(GL_TEXTURE_2D);

	glEnable(GL_DEPTH_TEST);

}